//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include <ffl.h>
#include "VPconfig.h"

using namespace std;

void PrintUsage(void){
  cerr<<endl;
  cerr<<"This command prints active data quality flag channels in a given time segment."<<endl;
  cerr<<"The ouput can be either a list of active flags or the list of time segments"<<endl;
  cerr<<"when the flags are active."<<endl;
  cerr<<endl;
  cerr<<"Usage:"<<endl;
  cerr<<endl;
  cerr<<"dq-print.exe gps=[GPS time] \\"<<endl;
  cerr<<"             gps-start=[GPS start] \\"<<endl;
  cerr<<"             gps-end=[GPS end] \\"<<endl;
  cerr<<"             dt=[half time window] \\"<<endl;
  cerr<<"             channel=[channel name pattern] \\"<<endl;
  cerr<<"             channel-file=[path to channel file] \\"<<endl;
  cerr<<"             ffl=[path to ffl] \\"<<endl;
  cerr<<"             active-value=[active value] \\"<<endl;
  cerr<<"             print=[print type] \\"<<endl;
  cerr<<"             output=[path to output file]"<<endl;
  cerr<<endl;
  cerr<<"[GPS time]             GPS time of interest"<<endl;
  cerr<<"[half time window]     half time window to be applied on both sides of the GPS time of interest [s]. Default dt=5"<<endl;
  cerr<<"[GPS start]            starting GPS time"<<endl;
  cerr<<"[GPS send]             ending GPS time"<<endl;
  cerr<<endl;
  cerr<<"[channel name pattern] filter channel names to consider. The filter string must be a pattern."<<endl;
  cerr<<"                       Default: channel=\"V1:DQ_*_FLAG_* V1:DQ*_VETO_*\""<<endl;
  cerr<<"[path to channel file] path to the file listing channel names."<<endl;
  cerr<<"                       they must be listed in a text file over a single column."<<endl;
  cerr<<"[path to ffl]          path to ffl (or lcf). Default = /virgoData/ffl/raw.ffl"<<endl;
  cerr<<"[active value]         value for the channel to be said \"active\". Default: val=1"<<endl;
  cerr<<endl;
  cerr<<"[print type]           \"segments\": print segments when at least one channel is active"<<endl;
  cerr<<"                       \"list-active\": print list of active channels"<<endl;
  cerr<<"                       \"list-all\": print list of all channels"<<endl;
  cerr<<"                       default: print=\"list-active\""<<endl;
  cerr<<"[path to output file]  path to output file. By default, use standard output."<<endl;
  cerr<<endl;
  return;
}
  
int main (int argc, char* argv[]){

  // number of arguments
  if(argc>1&&!((string)argv[1]).compare("version")){
    PrintVersion();
    return 0;
  }

  // number of arguments
  if(argc<2){
    PrintUsage();
    return 1;
  }

  // list of parameters + default
  string chfilter   ="V1:DQ_*_FLAG_* V1:DQ*_VETO_*"; // channel filter
  double gps_0      =-1.0;                     // GPS of interest
  double gps_start  =-1.0;                     // GPS start
  double gps_end    =-1.0;                     // GPS end
  double dt         =5.0;                      // time window
  string fflpath    ="/virgoData/ffl/raw.ffl"; // ffl
  double val        =1;                        // channel value
  string printtype  ="list-active";            // print type
  string outfile    ="";                       // output file
  string chfile     ="";                       // channel file

  // loop over arguments
  vector <string> sarg;
  for(int a=1; a<argc; a++){
    sarg=SplitString((string)argv[a],'=');
    if(sarg.size()!=2) continue;
    if(!sarg[0].compare("gps"))           gps_0=atof(sarg[1].c_str());
    if(!sarg[0].compare("gps-start"))     gps_start=atof(sarg[1].c_str());
    if(!sarg[0].compare("gps-end"))       gps_end=atof(sarg[1].c_str());
    if(!sarg[0].compare("dt"))            dt=atof(sarg[1].c_str());
    if(!sarg[0].compare("channel"))       chfilter=(string)sarg[1];
    if(!sarg[0].compare("channel-file"))  chfile=(string)sarg[1];
    if(!sarg[0].compare("ffl"))           fflpath=(string)sarg[1];
    if(!sarg[0].compare("active-value"))  val=atof(sarg[1].c_str());
    if(!sarg[0].compare("print"))         printtype=(string)sarg[1];
    if(!sarg[0].compare("output"))        outfile=(string)sarg[1];
  }

  // GPS of interest and window
  double gps_interest;
  if(gps_0>0){
    gps_interest=gps_0;
    if(gps_start<0&&gps_end<0&&dt>0){       gps_start=gps_0-dt;        gps_end=gps_0+dt; }
    else if(gps_start<0&&gps_end>0&&dt>0){  gps_start=gps_interest-dt; }
    else if(gps_start>0&&gps_end>0);
    else if(gps_start>0&&gps_end<0&&dt>0){                             gps_end=gps_interest+dt; }
    else{
      cerr<<"A valid gps timing must be provided"<<endl;
      cerr<<"Type 'dq-print' for help"<<endl;
      return 1;
    }
  }
  else{
    if(gps_start<0&&gps_end>0&&dt>0){ gps_interest=gps_end-dt; gps_start=gps_interest-dt; }
    else if(gps_start>0&&gps_end>0){ gps_interest=gps_start/2.0+gps_end/2.0; }
    else if(gps_start>0&&gps_end<0&&dt>0){ gps_interest=gps_start+dt; gps_end=gps_interest+dt; }
    else{
      cerr<<"A valid gps timing must be provided"<<endl;
      cerr<<"Type 'dq-print' for help"<<endl;
      return 1;
    }  
  }

  // check timing
  if(gps_interest<gps_start || gps_interest>gps_end){
    cerr<<"A valid gps timing must be provided"<<endl;
    cerr<<"Type 'dq-print' for help"<<endl;
    return 1;
  }  

  // print full command
  cout<<"Full command:"<<endl;
  cout<<"\tdq-print gps="<<setprecision(15)<<gps_interest<<" gps-start="<<gps_start<<" gps-end="<<gps_end<<" channel=\""<<chfilter<<"\" channel-file="<<chfile<<" ffl="<<fflpath<<" active-value="<<val<<" print="<<printtype<<" output="<<outfile<<endl;

  // ffl file
  ffl *ffldata = new ffl (fflpath);
  if(!ffldata->LoadFrameFile((int)gps_start)) return 2;

  // output file (print list only)
  ofstream ofs;
  if(outfile.compare("")&&printtype.compare("segments")){
    ofs.open(outfile,ofstream::out);
    if(!ofs.is_open()){
      cerr<<"Output file cannot be open: "<<outfile<<endl;
      return 3;
    }
  }

  // channel file (if any)
  if(chfile.compare("")&&IsTextFile(chfile)){
    string chname;
    ReadAscii *R = new ReadAscii(chfile, "s");
    for(int c=0; c<R->GetNRow(); c++){
      R->GetElement(chname, c, 0);
      chfilter+=" "+chname;
    }
    delete R;
  }
  
  // get channel list
  vector <string> channels;
  channels = ffldata->GetChannelList(chfilter);

  // print list of channels
  cout<<"Scanning "<<channels.size()<<" channels..."<<endl;
  for(int c=0; c<(int)channels.size(); c++) cout<<channels[c]<<" || ";
  cout<<endl;

  // loop over channels
  int chindex;
  int samplingrate;
  double chunkduration;
  int ns; double *vdata;
  double gps_s, gps_e, gps_c;
  double dt_min=-1e10;
  Segments *Seg = new Segments();
  for(int c=0; c<(int)channels.size(); c++){

    // get sampling rate
    samplingrate = ffldata->GetChannelSampling(channels[c],chindex);

    // missing channel
    if(samplingrate<=0){
      if(printtype.compare("segments")){
	if(outfile.compare("")) ofs<<"CH-MISSING \t "<<channels[c]<<endl;
	else cout<<"CH-MISSING \t "<<channels[c]<<endl;
      }
      continue;
    }
    
    // build and print segments
    if(!printtype.compare("segments")){

      // data chunk duration
      if(samplingrate<10.0) chunkduration=100000.0;
      else if(samplingrate<1000.0) chunkduration=1000.0;
      else if(samplingrate<10000.0) chunkduration=100.0;
      else chunkduration=10.0;
     
      // loop over data chunks
      gps_s=gps_start;
      while(gps_s<gps_end){
	gps_e=TMath::Min(gps_s+chunkduration,gps_end);
	
	// get data
	vdata=ffldata->GetData(ns,channels[c],floor(gps_s),ceil(gps_e));// get data vector
	if(ns<=0){ gps_s+=chunkduration; continue; }// no data

	// test data and fill segment list
	for(int d=0; d<ns; d++)
	  if(vdata[d]==val) Seg->AddSegment(floor(gps_s)+(double)d/(double)samplingrate, floor(gps_s)+(double)(d+1)/(double)samplingrate);

	// next
	gps_s+=chunkduration;
      }
    }

    // print channels
    else{
      
      // get data
      vdata=ffldata->GetData(ns,channels[c],floor(gps_start),ceil(gps_end));// get data vector
      if(ns<=0){// no data
	if(outfile.compare("")) ofs<<"CH-MISSING \t "<<channels[c]<<endl;
	else cout<<"CH-MISSING \t "<<channels[c]<<endl;
	continue;
      }
      dt_min=-1e10;

      // test channel value
      for(int d=0; d<ns; d++){
	gps_c=floor(gps_start)+((double)d+0.5)/(double)samplingrate;// current time
	if(vdata[d]==val&&TMath::Abs(gps_interest-gps_c)<TMath::Abs(dt_min)) // get min
	  dt_min=gps_c-gps_interest;
	if(dt_min>-1e10&&gps_c-gps_interest>-dt_min) break;
      }

      // channel is not active
      if(dt_min==-1e10){
	if(!printtype.compare("list-all")){
	  if(outfile.compare("")) ofs<<"NOT-ACTIVE \t "<<channels[c]<<endl;
	  else cout<<"NOT-ACTIVE \t "<<channels[c]<<endl;
	}
	delete vdata;
	continue;
      }      
      // check exact value
      if(TMath::Abs(dt_min)<=0.5/(double)samplingrate){
	if(outfile.compare("")) ofs<<showpos<<0.0<<"s \t "<<channels[c]<<endl;
	else cout<<showpos<<0.0<<"s \t "<<channels[c]<<endl;
	delete vdata;
	continue;
      }

      // print offset
      if(outfile.compare("")) ofs<<showpos<<dt_min<<"s \t "<<channels[c]<<endl;
      else cout<<showpos<<dt_min<<"s \t "<<channels[c]<<endl;
      delete vdata;

    }
  }

  // print segments
  if(!printtype.compare("segments")){
    
    // trim segments
    Seg->Intersect(gps_start, gps_end);
    
    // print segments
    if(outfile.compare("")) Seg->Write(outfile);
    else Seg->Dump();
  }
  else
    if(outfile.compare("")) ofs.close();
  
  cout<<"Channel scan is finished"<<endl;
  
  // cleaning
  channels.clear();
  delete Seg;
  delete ffldata;
  
  return 0;
}

